;;;;-*-lisp-*-
;;;; The Crypticl cryptographic library.
;;;;
;;;; Description: Pseudo random number generation.
;;;; Author: Taale Skogan <tasko@frisurf.no>
;;;; Distribution:  See the accompanying file LICENSE.


;;To do:
;;-get high entropy bits on non-Linux system. Either roll your own (most likely bad idea) or use win32API to handle one other system. But this is not important. win32 API CryptGenRandom.
;;-test suite. Some simple statistical tests on output?

(in-package crypticl)

(defun random-bignum (bitsize)
  "Return random integer bitsize bits long, i.e. an integer in the range [0, 2^bitsize - 1]. Uses the internal Lisp PSNG random which is _not_ cryptographically secure.
"
  (warn "Using a PSNG which is _not_ cryptographically secure.")
  (random-bignum-internal bitsize #'random))



(defun random-secure-bignum (bitsize)
  "Return random integer bitsize bits long generatated from a cryptograpically secure pseudo random number generator. The function is very slow because random-secure invokes SHA-1 multiple times. It should only be used for cryptographic keys and the like. 

Note that according to Menezes et al (1997), there exists no formal proof that this add-hoc solution using SHA-1 is cryptographically secure. But it is nevertheless approved for use in standards like the Digital Signature Standard (NIST 2000).
"
  (random-bignum-internal bitsize #'random-secure))


(defun random-bignum-internal (bitsize rand-function)
  "Internal version of random-bignum. rand-function is a random number generator which takes one integer argument r and returns an integer in the range [0,r-1]."
  (assert (> bitsize 0))	   
  (let* ((size (ceiling bitsize 8))
	 (octet-array (make-array size 
				  :element-type '(unsigned-byte 8)))
	 n)
    (do ((i 0 (1+ i)))
	((>= i size))
      (setf (aref octet-array i) (funcall rand-function 256)))
  
  ;;Remove extra bits if necessary. This is done by setting the 
  ;;unnecessary 8 - (bitsize mod 8) most significant bits to zero.
  (setf n (octet-vector-to-integer octet-array))
  (if (= 0 (mod bitsize 8))
      n
    (dpb 0  (byte (- 8 (mod bitsize 8)) bitsize) n))))


(defun random-bignum-max-odd (bitsize)
  "Return random, bitsize bits long, odd integer. In other words, the least and most significant bit is always 1."
  (let ((n (random-secure-bignum bitsize)))
    (setf n (dpb 1 (byte 1 (1- bitsize)) n)
	  n (dpb 1 (byte 1 0) n))))

	

;;;;;;;;
;;; CRYPTOGRAPHICALLY SECURE RANDOM NUMBER GENERATOR

(defparameter *random-secure-obj* nil
  "State for the random number generator")

(defun random-secure-range (low high)
  "Return random integer in the range [low,high]. This should only be used for values of low close to 0."
  (while t
    (let ((n (random-secure (+ high 1))))
      (when (>= n low)
	(return-from random-secure-range n)))))
      

(defun random-secure (n)
  "Return random integer in the range [0,n-1]."
  (if (not *random-secure-obj*)
      (setf *random-secure-obj* (make-SecurePRNG)))
  (SecurePRNG-random *random-secure-obj* n))

(defun random-secure-octets (size)
  "Returns size pseudorandom octets from a cryptographically secure PRNG."
  (if (not *random-secure-obj*)
      (setf *random-secure-obj* (make-SecurePRNG)))
  (SecurePRNG-octets *random-secure-obj* size))

(defclass SecurePRNG ()
  ((seed :accessor seed
	 :initarg :seed))
  (:documentation "Cryptographically secure pseudo random number generator."))

(defun set-seed (seed)
  "Creates a new seed for the secure PRNG. Input should be a high entropy bignum at least 160 bits long."
  (if (not *random-secure-obj*)
      (setf *random-secure-obj* (make-SecurePRNG)))
  (setf (seed *random-secure-obj*) seed)) 

(defun make-SecurePRNG ()
  "Constructor for the Secure-PRNG class. Assumes that a 160 bits secret/seed is enough."
  (make-instance 'SecurePRNG
    :seed (random-secure-seed 160)))


(defmethod SecurePRNG-random ((obj SecurePRNG) n)
  "Return random number in the range [0,n-1] where n <= 2^160. Is terrible inefficient for small n because we waste most of the 160 bits SHA-1 returns."
  (assert (<= (integer-length n) 160)) 
  (let* ((seed (seed obj))
	 (state (octet-vector-to-integer
		 (sha1-on-octet-vector (integer-to-octet-vector seed)))))
    
    ;;update seed to get forward security
    (setf (seed obj) (mod (+ 1 seed state)
			  (expt 2 160)))
    
    ;;reduce n to proper size
    (mod state n)))


(defmethod SecurePRNG-octets ((obj SecurePRNG) size)
  "Returns size pseudorandom octets from a cryptographically secure PRNG."
  (let ((rounds (ceiling size 20))	;20 octets in 160 bits
	(ret (make-array size :element-type '(unsigned-byte 8)))
	(current-size 0)		;num octets processed so far
	(hash nil))
    
    (dotimes (i rounds)
      (setf hash (sha1-on-octet-vector (integer-to-octet-vector (seed obj))))    
      (do ((j 0 (1+ j)))
	  ((or (>= current-size size) (>= j 20)))
		  
	(setf (aref ret current-size) (aref hash j))
	(incf current-size))
      	    
      ;;update seed to get forward security
      (setf (seed obj) (mod (+ 1 (seed obj) (octet-vector-to-integer hash))
			    (expt 2 160))))
    ret))
    


(defun high-entropy-octets (size)
  "Return size octets from some hopefully high entropy bit source."
  (let ((ret (make-array size :element-type '(unsigned-byte 8))))
    (with-open-file (file "/dev/random" :direction :input
		     :if-does-not-exist nil)          
      (if (not file)
	  (progn
		(warn "/dev/random was not available on this system.")
		(warn "Add 160 bits high entropy seed maually using set-seed.")
		(warn "Will use random in the mean time.")
		(do ((i 0 (1+ i)))
		    ((>= i size))
		  (setf (aref ret i) (random 256))))	
	
	(progn
	  (do ((i 0 (1+ i)))
		  ((>= i size))
	    ;;Fails if /dev/random runs out of bytes, but that 
	    ;;should never happen.
	    (setf (aref ret i) (read-byte file nil)))))
      ret)))

  

(defun random-secure-seed (bitsize)
  "Returns bitsize integer with full entrophy (enthropy equals bitsize). Only works on Linux-like systems where /dev/random is a source of high enthropy bits."
  (let* ((size (ceiling bitsize 8))
	 (octet-array (high-entropy-octets size))
	 n)				;Return value
      
    ;;Remove extra bits if necessary. This is done by setting the 
    ;;unnecessary 8 - (bitsize mod 8) most significant bits to zero.
    (setf n (octet-vector-to-integer octet-array))
    (if (= 0 (mod bitsize 8))
	n
      (dpb 0  (byte (- 8 (mod bitsize 8)) bitsize) n))))
  